(* Copyright 2009, Thant Tessman. Permission is granted for
   non-commercial use provided this copyright notice is preserved. All
   other rights reserved. *)

structure Tokens = Tokens

type pos = int
type svalue = Tokens.svalue
type ('a,'b) token = ('a,'b) Tokens.token
type lexresult = (svalue,pos) token

val ln = DL.lineno

val eof = fn () => Tokens.EOF(!ln, !ln)

val prefix = DL.prefix

val error = fn e => TextIO.output(TextIO.stdErr, 
				  !prefix 
				  ^ "lex error line " 
				  ^ Int.toString(!ln) 
				  ^ ": " ^ e ^ "\n")

%%

%header (functor DLLexerFun(structure Tokens: DLGrammar_TOKENS));

alpha = [a-zA-Z_];
digit = [0-9];
id = {alpha}({alpha}|{digit})*;
symbol = #{id};

ws = [\ \t];
nat = [0-9]+;
real = (([0-9]+(\.[0-9]+)?)|(\.[0-9]+))([eE][+-]?[0-9]+)?;

path = "$"{id}((\.{id})|(\[{nat}\]))*;

vecn = "vec"{nat};

escseq = \\([abtnvfr\\\"\']|[0-7]{3}|("x"[0-9a-fA-F]{2}));

validchar = [^\n\\\"\']|{escseq};

char = \'({validchar}|\")\';

string = \"({validchar}|\')*\";

%%

"\n" 	=>	(ln := !ln + 1; lex());
{ws}+	=>	(lex());

{path}	=>	(Tokens.PATH (ParsePath.parse yytext, !ln, !ln));

"any"	=>	(Tokens.ANY_LIT(!ln, !ln));

"char"	=>	(Tokens.CHR_LIT(!ln, !ln));
"enum"	=>	(Tokens.ENUM_LIT(!ln, !ln));
"int"	=>	(Tokens.INT_LIT(!ln, !ln));
"real"	=>	(Tokens.REAL_LIT(!ln, !ln));
"sym"	=>	(Tokens.SYM_LIT(!ln, !ln));

"vec"	=>	(Tokens.VEC_LIT(!ln, !ln));

{vecn}	=>	(let
		     val s = String.extract(yytext, 3, NONE);
		 in
		     case Int.fromString(s) of
			 SOME i => Tokens.VECN_LIT(i, !ln, !ln)
		       | NONE => (error("bad vecN: " ^ yytext);
				  Tokens.VECN_LIT(0, !ln, !ln))
		 end);

"rec"	=>	(Tokens.REC_LIT(!ln, !ln));
"type"	=>	(Tokens.TYPE_LIT(!ln, !ln));

"{"	=>	(Tokens.LEFT_CURLY(!ln, !ln));
"}"	=>	(Tokens.RIGHT_CURLY(!ln, !ln));

"["	=>	(Tokens.LEFT_SQUARE(!ln, !ln));
"]"	=>	(Tokens.RIGHT_SQUARE(!ln, !ln));

":"	=>	(Tokens.COLON(!ln, !ln));
"="	=>	(Tokens.EQUAL(!ln, !ln));
","	=>	(Tokens.COMMA(!ln, !ln));
"-"	=>	(Tokens.MINUS(!ln, !ln));

{char}	=>	(let
		     val l = String.size(yytext)
		     val s = String.extract(yytext, 1, SOME (l-2))
		 in
		     case Char.fromCString(s) of
			 SOME c => Tokens.CHR(c, !ln, !ln)
		       | NONE => (error ("bad char: " ^ yytext);
				  Tokens.CHR(chr 0, !ln, !ln))
		 end);

{nat}   =>      (case Int.fromString(yytext) of
		   SOME n => Tokens.NAT(n, !ln, !ln)
		 | NONE => (error("bad nat: " ^ yytext);
			      Tokens.NAT(0, !ln, !ln)));

{real}	=>	(case Real.fromString(yytext) of
		     SOME r => Tokens.REAL(r, !ln, !ln)
		   | NONE => (error("bad real: " ^ yytext);
			      Tokens.REAL(0.0, !ln, !ln)));

{string} =>	(let
		     val l = String.size(yytext)
		     val c_s = String.extract(yytext, 1, SOME (l-2))
		 in
		     case String.fromCString(c_s) of
			 SOME s => Tokens.STR(s, !ln, !ln)
		       | NONE => (error("bad string: " ^ yytext);
				  Tokens.STR("", !ln, !ln))
		 end);

{symbol} =>	(let
		     val s = String.extract(yytext, 1, NONE);
		 in
		     Tokens.SYM(Path.stringToSymbol(s), !ln, !ln)
		 end);

{id}	=>	(Tokens.NAME(Path.stringToSymbol(yytext), !ln, !ln));

.	=>	(error ("lex error: "^yytext); lex());
