(* Copyright 2009, Thant Tessman. Permission is granted for
   non-commercial use provided this copyright notice is preserved. All
   other rights reserved. *)

structure DL = 
struct

val lineno = ref 1
val prefix = ref ""

fun reset () = (lineno := 1; prefix := "")

type path = Path.path
type sym = Path.sym
type elem = Path.elem

val pathToString = Path.pathToString
val symbolToString = Path.symbolToString
val sameSym = Path.sameSym
val samePath = Path.samePath
val sameElem = Path.sameElem


datatype dltype 
  = ANY_TYPE
  | CHR_TYPE
  | ENUM_TYPE of sym list
  | INT_TYPE
  | NO_TYPE (* most general type *)
  | REAL_TYPE
  | REC_TYPE of (sym * dltype) list
  | SYM_TYPE
  | TYPE_PATH of path
  | TYPE_REF of dltype ref
  | VEC_TYPE of dltype
  | VECN_TYPE of (int * dltype)


datatype dlbinding 
  = BINDING of (sym * dltype * dldata)
  | TYPEDEF of (sym * dltype)

and dldata 
  = CHR of char
  | DATA_PATH of path
  | DATA_REF of dldata ref
  | INT of int
  | REAL of real
  | REC of dlbinding list
  | STR of string
  | SYM of sym
  | VEC of dldata list

datatype value
  = DATA of dldata
  | TYPE of dltype

type rec_type = (sym * dltype) list

fun dumpValue (stream, value) = 
    let
      fun dump text = TextIO.output(stream, text)
      fun dumpl text_list = List.app dump text_list

      val tab = "  "

      fun indent i = 
	  if i>0
	  then (dump tab; indent (i-1))
	  else ()

      fun dumpBindingTypes ([], i) = ()
	| dumpBindingTypes ((s,t) :: rest, i) = 
	  (dumpBindingTypes (rest, i)
	 ; indent i
	 ; dumpl [(symbolToString s), " : "]
         ; dumpType (t, i)
	 ; dump "\n")

      and dumpEnums ([], _) = ()
	| dumpEnums (s :: rest, i) =
	  (dumpEnums (rest, i)
	 ; indent i
	 ; dumpl ["#", symbolToString s, "\n"])


      and dumpType (t, i) =
	  case t of
	    ANY_TYPE => dump "any"
	  | NO_TYPE => dump "none"
	  | CHR_TYPE => dump "char"
	  | INT_TYPE => dump "int"
	  | REAL_TYPE => dump "real"
	  | ENUM_TYPE enums => 
	    (dump "enum {\n"
	   ; dumpEnums (enums, i+1)
	   ; indent i
	   ; dump "}")
	  | REC_TYPE bindings =>
	    (dump "rec {\n"
	   ; dumpBindingTypes (bindings, i+1)
	   ; indent i
	   ; dump "}")
	  | SYM_TYPE => dump "sym"
	  | TYPE_PATH p => dumpl ["$", pathToString p]
	  | TYPE_REF (ref t') => dumpType(t', i)
(*	  | TYPE_REF (ref t') => (dump " ref "; dumpType(t', i)) *)
	  | VEC_TYPE vt => (dump "vec "; dumpType (vt, i))
	  | VECN_TYPE (n, vt) =>
	    (dumpl ["vec", Int.toString(n), " "]
	   ; dumpType (vt, i))

      and dumpBindings ([], _) = ()
	| dumpBindings ((BINDING (s,t,d))::rest, i) =
	  (dumpBindings(rest, i)
	 ; dump "\n"
	 ; indent i
	 ; dump (symbolToString s)
	 ; case t of 
	     ANY_TYPE => ()
	   | _ => (dump " : "; dumpType (t, i))
	 ; dump " = "
	 ; dumpData (d, i))
	| dumpBindings ((TYPEDEF (s,t))::rest, i) =
	  (dumpBindings (rest, i)
	 ; dump "\n"
	 ; indent i
	 ; dumpl ["type ", symbolToString s, " = "]
	 ; dumpType (t, i))

      and dumpElements ([], _) = ()
	| dumpElements (element::rest, i) =
	  (dump "\n"
	 ; indent i
	 ; dumpData (element, i)
	 ; dump ","
	 ; dumpElements (rest, i))

      and dumpData (d, i) =
	  case d of 
	    CHR c => dumpl ["'", Char.toCString c, "'"]
	  | DATA_PATH p => dumpl ["$", pathToString p]
	  | DATA_REF (ref d') => dumpData(d', i)
(*	  | DATA_REF (ref d') => (dump " ref "; dumpData(d', i)) *)
	  | INT i => if i<0 
		     then dumpl ["-", Int.toString (~i)]
		     else dump (Int.toString i)
	  | REAL r => if Real.signBit r 
		      then dumpl ["-", Real.toString (~r)]
		      else dump (Real.toString r)
	  | SYM sym => dumpl ["#", symbolToString sym]
	  | STR s => dumpl ["\"", String.toCString s, "\""]
	  | VEC elements => (dump "["
			   ; dumpElements(elements,i+1)
			   ; dump "\n"
			   ; indent i
			   ; dump "]")
	  | REC bindings => (dump "{"
			   ; dumpBindings (bindings,i+1)
			   ; dump "\n"
			   ; indent i
			   ; dump "}")

    in
      case value of
	DATA d =>
	(case d of 
	   REC bindings => (dumpBindings(bindings, 0)
			  ; dump "\n")
	 | _ => dumpData(d, 0))
      | TYPE t => dumpType (t, 0)
    end


(* Find the common type of two types. The isa function won't suffice
   in the case of records. *)

local

  fun mergeSym (a, b_list) =
      case List.find (fn(b) => sameSym (a,b)) b_list of
	SOME _ => b_list
      | NONE => a :: b_list

  fun mergeEnum (a :: a_list, b_list) =
      mergeSym(a, mergeEnum(a_list, b_list))
    | mergeEnum ([], b_list) = b_list

in

fun commonType tt =
    case tt of

      (NO_TYPE, b) => b
    | (a, NO_TYPE) => a

    | (CHR_TYPE, CHR_TYPE) => CHR_TYPE
    | (INT_TYPE, INT_TYPE) => INT_TYPE

    | (INT_TYPE, REAL_TYPE) => REAL_TYPE
    | (REAL_TYPE, INT_TYPE) => REAL_TYPE

    | (REAL_TYPE, REAL_TYPE) => REAL_TYPE

    | (SYM_TYPE, SYM_TYPE) => SYM_TYPE

    | (SYM_TYPE, ENUM_TYPE _) => SYM_TYPE
    | (ENUM_TYPE _, SYM_TYPE) => SYM_TYPE


    | (ENUM_TYPE a_list, ENUM_TYPE b_list) 
      => ENUM_TYPE (mergeEnum (a_list, b_list))

    | (TYPE_REF (ref a), b) => commonType (a, b)
    | (a, TYPE_REF (ref b)) => commonType (a, b)

    | (VEC_TYPE a, VEC_TYPE b) => VEC_TYPE (commonType (a, b))

    | (VECN_TYPE (_,a), VEC_TYPE b) => VEC_TYPE (commonType (a, b))
    | (VEC_TYPE a, VECN_TYPE (_,b)) => VEC_TYPE (commonType (a, b))

    | (VECN_TYPE (size_a,a), VECN_TYPE (size_b,b)) =>
      (if size_a=size_b
       then VECN_TYPE (size_a, commonType (a, b))
       else VEC_TYPE (commonType (a, b)))

    | (REC_TYPE a_bts, REC_TYPE b_bts) =>
      let
	fun findCommonBindings [] = []
	  | findCommonBindings ((b_s,b_t)::b_rest) =
	    let
	      fun findCommonBinding [] = NONE
		| findCommonBinding ((a_s, a_t)::a_rest) =
		  if sameSym (a_s, b_s) 
		  then SOME (a_s, commonType (a_t, b_t))
		  else findCommonBinding a_rest
	    in
	      case findCommonBinding a_bts of
		SOME bt => bt :: findCommonBindings b_rest
	      | NONE => findCommonBindings b_rest
	    end
      in
	REC_TYPE (findCommonBindings b_bts)
      end

    | (_, _) => ANY_TYPE

end


(* Find the most specific type of two types. If two types don't
   overlap the result is NO_TYPE. *)

local

structure SymKey = 
struct 
  type key = sym 
  val compare = Path.compareSym
end

structure SymDict = Dict(SymKey)

in

fun specificType tt =
    let

      (* Find the set of symbols common to both lists. *)

      fun commonSyms ([], _) = []
        | commonSyms (a :: a_rest, b_list) = 
          case List.find (fn(b) => sameSym(a,b)) b_list of 
            SOME _ => a :: commonSyms(a_rest, b_list)
          | NONE => commonSyms(a_rest, b_list)

      fun concatRecTypes(a_bts, b_bts) =
          let

            fun buildDict ([], dict) = dict
              | buildDict ((s, t) :: rest, dict) 
                = buildDict (rest, SymDict.insert(dict, s, t))

            fun extendDict ([], dict) = dict
              | extendDict ((s, t) :: rest, dict) =
                case SymDict.find (dict, s) of
                  NONE => extendDict(rest, SymDict.insert(dict, s, t))
                | SOME t' => extendDict(rest, SymDict.insert(dict, s, specificType(t, t')))

            val dict = buildDict(a_bts, SymDict.newDict())
            val dict' = extendDict(b_bts, dict)

          in

            SymDict.toList(dict')

          end
    in

      case tt of

        (ANY_TYPE, t) => t
      | (t, ANY_TYPE) => t

      | (REC_TYPE a_bts, REC_TYPE b_bts) => (REC_TYPE (concatRecTypes(a_bts, b_bts)))

      | (VECN_TYPE (size_a,a), VECN_TYPE (size_b,b)) =>
        (if size_a=size_b
         then VECN_TYPE (size_a, specificType (a, b))
         else NO_TYPE)

      | (VECN_TYPE (size,a), VEC_TYPE b) => VECN_TYPE (size, specificType (a, b))
      | (VEC_TYPE a, VECN_TYPE (size,b)) => VECN_TYPE (size, specificType (a, b))
            
      | (VEC_TYPE a, VEC_TYPE b) => VEC_TYPE (specificType (a, b))

      | (TYPE_REF (ref a), b) => specificType (a, b)
      | (a, TYPE_REF (ref b)) => specificType (a, b)

      | (ENUM_TYPE a_list, ENUM_TYPE b_list) 
        => ENUM_TYPE (commonSyms (a_list, b_list))

      | (SYM_TYPE, ENUM_TYPE enums) => ENUM_TYPE enums
      | (ENUM_TYPE enums, SYM_TYPE) => ENUM_TYPE enums

      | (SYM_TYPE, SYM_TYPE) => SYM_TYPE

      | (REAL_TYPE, REAL_TYPE) => REAL_TYPE

      | (INT_TYPE, REAL_TYPE) => INT_TYPE
      | (REAL_TYPE, INT_TYPE) => INT_TYPE

      | (CHR_TYPE, CHR_TYPE) => CHR_TYPE
      | (INT_TYPE, INT_TYPE) => INT_TYPE

      | (_, _) => NO_TYPE

    end
end

local

  fun checkSym (a, []) = false
    | checkSym (a, b::rest) = if sameSym (a,b) 
			      then true 
			      else checkSym (a,rest)

  fun checkEnum ([], _) = true
    | checkEnum (a :: a_rest, b_list) = 
      checkSym (a, b_list) andalso checkEnum (a_rest, b_list)

in

fun isa tt =
    case tt of

      (_, ANY_TYPE) => true

    | (CHR_TYPE, CHR_TYPE) => true

    | (INT_TYPE, INT_TYPE) => true
    | (INT_TYPE, REAL_TYPE) => true
    | (REAL_TYPE, REAL_TYPE) => true

    | (SYM_TYPE, SYM_TYPE) => true
    | (ENUM_TYPE _, SYM_TYPE) => true

    
    | (ENUM_TYPE a_list, ENUM_TYPE b_list) => checkEnum (a_list, b_list)

    | (TYPE_REF (ref a), b) => isa (a, b)
    | (a, TYPE_REF (ref b)) => isa (a, b)

    | (VEC_TYPE a, VEC_TYPE b) => isa (a, b)
    | (VECN_TYPE (_,a), VEC_TYPE b) => isa (a, b)
    | (VECN_TYPE (a_size, a), VECN_TYPE (b_size, b)) 
      => (a_size = b_size) andalso isa (a, b)

    | (REC_TYPE a_bts, REC_TYPE b_bts) =>
      let
	fun checkBindingTypes ((b_s,b_t)::b_rest) =
	    let
	      fun checkBindingType [] = false
		| checkBindingType ((a_s, a_t)::a_rest) = 
		  if sameSym (a_s, b_s) 
		  then isa (a_t, b_t)
		  else checkBindingType a_rest
	    in
	      (checkBindingType a_bts) andalso checkBindingTypes b_rest
	    end
	  | checkBindingTypes [] = true
      in
	checkBindingTypes b_bts
      end

    | (_,_) => false
end


(* fun getType (data) *)

local 
  fun getDataType d =
      case d of 
	CHR _ => CHR_TYPE
      | DATA_PATH _ => ANY_TYPE
      | DATA_REF (ref d) => getDataType d
      | INT _ => INT_TYPE
      | REAL _ => REAL_TYPE
      | SYM s => ENUM_TYPE [s]
      | REC bindings => REC_TYPE (getBindingTypes bindings)
      | STR s => VECN_TYPE (String.size s, CHR_TYPE)
      | VEC v => VECN_TYPE (List.length v, getElementType v)

  and getBindingTypes ((BINDING (s, _, d)) :: rest) =
      (s, getDataType d) :: (getBindingTypes rest)
    | getBindingTypes ((TYPEDEF _) :: rest) = (getBindingTypes rest)
    | getBindingTypes [] = []

  and getElementType (elem::rest) = 
      commonType(getDataType elem, getElementType rest)
    | getElementType [] = NO_TYPE

in
fun getType data = getDataType data
end


(* A type check error is (path * expected * found). *)

local
  fun checkBindings (bindings, errors, sofar) =
      case bindings of
	(TYPEDEF _)::rest => checkBindings (rest, errors, sofar)
      | (BINDING (s,a_t,d))::rest => 
	let
	  val b_t = getType d
	  val sofar' = (Path.BINDING s)::sofar
	  val errors' = checkData (d, errors, sofar')
	  val errors'' = checkBindings (rest, errors', sofar)
	in
	  if isa (b_t, a_t) 
	  then errors''
	  else (rev sofar', a_t, b_t) :: errors''
	end
      | [] => errors

  and checkElements (i, elems, errors, sofar) = 
      case elems of
	elem::rest =>
	let
	  val sofar' = (Path.INDEX i)::sofar
	  val errors' = checkData (elem, errors, sofar')
	in
	  checkElements (i+1, rest, errors', sofar)
	end
      | [] => errors

  and checkData (data, errors, path) =
      case data of
	DATA_REF (ref data') => checkData (data', errors, path)
      | REC bindings => checkBindings (bindings, errors, path)
      | VEC elems => checkElements (0, elems, errors, path)
      | _ => errors

in
fun checkTypes data = checkData(data, [], [])
end

fun dumpTypeErrors (stream, errors) =
    let
      fun dump text = TextIO.output(stream, text);
      fun dumpl text_list = List.app dump text_list

      fun dumpErrors ((path, expected, found) :: rest) =
	  (dumpErrors rest
	 ; dumpl ["\ntype error at: $", pathToString path]
	 ; case expected of
	     TYPE_PATH _ => dump "\ntype is unresolved path: "
	   | _ => dump "\nexpected: "
	 ; dumpValue (stream, TYPE expected)
	 ; case found of
	     TYPE_PATH _ => dump "\nvalue is unresolved path: "
	   | _ => dump "\nfound: "
	 ; dumpValue (stream, TYPE found)
	 ; dump "\n")
      | dumpErrors [] = ()
    in
      dumpErrors errors
    end

(* findDuplicates (dldata) -> path list *)


(* In theory, type bindings and data bindings could be considered
separate name spaces. But I think this would be unnecessarily
confusing. *)

local

  fun isUnique (_, []) = true
    | isUnique (s1, (TYPEDEF (s2, _)) :: rest) 
      = not (sameSym (s1, s2)) andalso isUnique (s1, rest)
    | isUnique (s1, (BINDING (s2, _, _)) :: rest)
      = not (sameSym (s1, s2)) andalso isUnique (s1, rest)

  fun checkBindings (bindings, sofar) =
      case bindings of
	(BINDING (sym, _, data)) :: rest => 
	let
	  val sofar' = (Path.BINDING sym) :: sofar
	  val errors = checkData (data, sofar') 
		       @ checkBindings (rest, sofar)
	in
	  if isUnique (sym, rest)
	  then errors
	  else (rev sofar') :: errors
	end
      | (TYPEDEF (sym, _)) :: rest => 
	let
	  val sofar' = (Path.BINDING sym) :: sofar
	  val errors = checkBindings (rest, sofar)
	in
	  if isUnique (sym, rest)
	  then errors
	  else (rev sofar') :: errors
	end
      | [] => []

  and checkElements (index, elems, sofar) =
      case elems of 
	elem :: rest => 
	let
	  val sofar' = (Path.INDEX index) :: sofar
	  val errors = checkData (elem, sofar')
		       @ checkElements (index + 1, rest, sofar)
	in
	  errors
	end
      | [] => []

  and checkData (data, sofar) =
      case data of
	DATA_REF (ref data') => checkData (data', sofar)
      | REC bindings => checkBindings (bindings, sofar)
      | VEC elements => checkElements (0, elements, sofar)
      | _ => []
in
fun findDuplicates (data) = checkData (data, [])
end


fun dumpDuplicateErrors (stream, errors) =
    case errors of
      [] => ()
    | _ =>
      let
	fun dump text = TextIO.output (stream, text)
	fun dumpl text_list = List.app dump text_list
	fun dumpPath path = dumpl ["$", pathToString path, "\n"]
      in
	(dump "\nduplicate binding name(s): \n"
       ; List.app dumpPath errors)
      end


datatype path_tag
  = NOT_A_RECORD
  | NOT_A_VECTOR
  | BAD_INDEX
  | BAD_BINDING
  | VALID_PATH
  | NOT_FOUND

type location = {path: path, 
                 tag: path_tag,
                 value: value option}

fun pathTagToString (tag : path_tag) =
    case tag of
      NOT_A_RECORD =>    "not a record"
    | NOT_A_VECTOR =>    "not a vector"
    | BAD_INDEX =>       "invalid index"
    | BAD_BINDING =>     "bad binding"
    | NOT_FOUND =>       "not found"
    | VALID_PATH =>      "valid path"


(* fun checkPath(path,data) -> location *)

local

  val rec doBindings : sym * path * path * dlbinding list -> location

    = fn (sym1, path, sofar, binding :: bindings) =>

         (case binding of

            BINDING (sym2, _, data) =>
            if sameSym (sym1, sym2)
            then doData(data, path, (Path.BINDING sym1) :: sofar)
            else doBindings(sym1, path, sofar, bindings)

          | TYPEDEF (sym2, t) =>
            if sameSym (sym1, sym2) 
            then
              let
                val here = rev ((Path.BINDING sym1) :: sofar)
              in
                case path of 
                  [] => {path = here,
                         tag = VALID_PATH,
                         value = SOME (TYPE t)}
                | (Path.BINDING _) :: _ => {path = here,
                                            tag = NOT_A_RECORD,
                                            value = NONE}
                | (Path.INDEX _) :: _ => {path = here,
                                          tag = NOT_A_VECTOR,
                                          value = NONE}
              end
           else doBindings(sym1, path, sofar, bindings))

       | (_, _, sofar, []) => 
         {path = rev sofar,
          value = NONE,
          tag = case sofar of [] => NOT_FOUND | _ => BAD_BINDING}


  (* data, rest of path, path so far (reversed) *)

  and doData: dldata * path * path -> location = 

   fn (data, [], sofar) => {path = rev sofar, 
                            tag = VALID_PATH, 
                            value = SOME (DATA data)}



    | (DATA_REF dr, path, sofar) => doData(!dr, path, sofar)

    | (data, path as elem::elems, sofar) =>
      case elem of
	Path.BINDING sym1 =>
	(case data of
	   REC bindings => doBindings (sym1, elems, sofar, bindings)
	 | _ => {path = rev sofar, 
                 tag = NOT_A_RECORD,
                 value = NONE})
      | Path.INDEX i =>
	(case data of
	   VEC items => (doData (List.nth (items, i),
				 elems,
				 elem :: sofar)
	                 handle Subscript => {path = rev (elem :: sofar),
                                              tag = BAD_INDEX,
                                              value = NONE})
	 | _ => {path = rev sofar, 
                 tag = NOT_A_VECTOR,
                 value = NONE})
in
fun checkPath (path, data) = doData(data, path, [])
end


(* As the findAbsPath function descends into the data structure, it
   keeps track of a stack of scopes. Each scope contains the currently
   bound values. This is what is checked to see if a given reference
   path is valid.

   Since paths can only begin with symbols (which was just a design
   decision), the stack only consists of the bindings of the enclosing
   records and skips enclosing vectors. This means the path can't be
   inferred from the stack. So the path to the given scope in the
   stack is a complete path.

   Since path elements are added and removed freqently on the end of
   the current path as the function descends in the data structure,
   the path is stored backwards for efficiency. *)


datatype ref_tag
  = VALID_TYPE_REF of path
  | VALID_DATA_REF of path
  | EXPECTED_DATA of path
  | EXPECTED_TYPE of path
  | BAD_REF of location

type ref_error = {at: path, tag: ref_tag}


(* The findRefs function returns a list of ref_errors which contains
   every reference that occurrs in the data, but the reference is
   resolved into an absolute path (when it can be resolved). The
   tagged ref also contains the absolute path where the reference
   occurred.

   This is all information used first to report reference errors, and
   second to allow the 'addIndirection' function to replace every data
   and type binding referenced by a path with an SML-style
   reference. This way, when the paths are actually resolved, they are
   resolved to (SML-style) references.

   findRefs(dldata) -> ref_error list 

*)


local

  type binding_types = ((sym * dltype) list)
  type dlbindings = dlbinding list
  type scope = path * dlbindings
  type stack = scope list

  fun findAbsPath (path, (rev_scope_path, bindings) :: rest) = 
      let
        val tagged_path = checkPath(path, REC bindings)
        val tag = #tag tagged_path
        val value = #value tagged_path
      in
        case tag of
          NOT_FOUND => findAbsPath(path, rest) (* try up a scope *)
        | _ => {path = (rev rev_scope_path) @ (#path tagged_path),
                tag = tag,
                value = value}
      end
    (* hit top of stack *)
    | findAbsPath (path, []) = {path = path, tag = NOT_FOUND, value = NONE}

  val rec doBindingTypes : stack * binding_types * path -> ref_error list
    = fn (stack, (name, t) :: rest, sofar) =>
	 doType(stack, t, (Path.BINDING name) :: sofar) @ 
	 doBindingTypes(stack, rest, sofar)
       | (_, [], _) => []

  and doType : stack * dltype * path -> ref_error list
    = fn (stack, t, sofar) =>
	 case t of 
	   TYPE_PATH path => 
           let 
             val ref_error = findAbsPath (path, stack)
           in
             (case ref_error of
                {path = path', tag = VALID_PATH, value = SOME (TYPE _)} =>
                {tag = VALID_TYPE_REF path', at = rev sofar}
              | {path = path', tag = VALID_PATH, ...} =>
                {tag = EXPECTED_TYPE path', at =rev sofar}
              | _ => {tag = BAD_REF ref_error, at = rev sofar})
             :: []
           end
	 | REC_TYPE binding_types => 
	   doBindingTypes (stack, binding_types, sofar)
	 | VEC_TYPE t' => doType (stack, t', sofar)
	 | VECN_TYPE (_, t') => doType (stack, t', sofar)
	 | TYPE_REF tr => doType (stack, !tr, sofar)
	 | _ => []

  val rec doBindings : scope * stack * dlbindings * path -> ref_error list

    = fn (scope, stack, binding :: rest, sofar) =>
	 let
	   val (scope_path, bound) = scope
	 in
	   case binding of
	     BINDING (name, t, data) =>
	     let
	       val sofar' = Path.BINDING name :: sofar
	     in
	       doType(scope :: stack, t, sofar') 
	       @ doData(scope :: stack, data, sofar') 
	       @ doBindings((scope_path, binding :: bound), stack, rest, sofar)
	     end
	   | TYPEDEF (name, t) =>
	     let
	       val sofar' = (Path.BINDING name) :: sofar
	     in
	       doType(scope :: stack, t, sofar') 
	       @ doBindings((scope_path, binding :: bound), stack, rest, sofar)
	     end
	 end
       | (_, _, [], _) => []

  and doElements : int * stack * dldata list * path -> ref_error list
    = fn (i, stack, data :: rest, sofar) =>
	 doData(stack, data, (Path.INDEX i) :: sofar) 
	 @ doElements(i+1, stack, rest, sofar)
       | (_, _, [], _) => []

  and doData : stack * dldata * path -> ref_error list 
    = fn (stack, data, sofar) =>
	 case data of 
           DATA_PATH path =>
	   let 
             val tagged_path = findAbsPath(path, stack)
             val here = rev sofar
	   in
	     (case tagged_path of
                {path = path', tag = VALID_PATH, value = SOME (DATA _)} =>
                {tag = VALID_DATA_REF path', at = here}
              | {path = path', tag = VALID_PATH, value = SOME (TYPE _)} =>
                {tag = EXPECTED_DATA path', at = here}
	      | _ => {tag = BAD_REF tagged_path, at = here})
             :: []
	   end
	 | REC bindings => doBindings ((sofar, []), stack, rev bindings, sofar)
	 | VEC elements => doElements (0, stack, elements, sofar)
	 | DATA_REF dr => doData (stack, !dr, sofar)
	 | _ => []

in
fun findRefs(data) = doData([], data, [])
end


fun dumpRefErrors (stream, errors) = 
    let
      fun dump text = TextIO.output(stream, text);
      fun dumpl text_list = List.app dump text_list
                            
      fun dumpErrors ({at, tag} :: rest) =
	  (case tag of 
	     BAD_REF bad_ref => 
 	     dumpl ["\nbad ref at: $", pathToString at,
                    "\n", pathTagToString (#tag bad_ref),
                    ": $", pathToString (#path bad_ref), "\n"]
	   | EXPECTED_TYPE path => 
             dumpl ["\nexpected ref to type at: $",
	            pathToString at,
	            "\ninstead found ref to data: $",
	            pathToString path, "\n"]
	   | EXPECTED_DATA path => 
	     dumpl ["\nexpected ref to data at: $",
	            pathToString at,
	            "\ninstead found ref to type: $",
	            pathToString path, "\n"]
	   | _ => ()
         ; dumpErrors rest)
        | dumpErrors([]) = ()
  in
    dumpErrors errors
  end



val addIndirections : (dldata * ref_error list) -> dldata = 
 fn (data, refs) =>
    let

      val rec buildTypePathDict : ref_error list -> (elem, unit) Tree.node
        = fn ({tag = VALID_TYPE_REF path, ...} :: rest) =>
             Tree.insert (sameElem, path, (), buildTypePathDict rest)
           | (_ :: rest) => buildTypePathDict rest 
           | [] => Tree.NODE (NONE, [])

      val rec buildDataPathDict : ref_error list -> (elem, unit) Tree.node
        = fn ({tag = VALID_DATA_REF path, ...} :: rest) =>
             Tree.insert (sameElem, path, (), buildDataPathDict rest)
           | (_ :: rest) => buildDataPathDict rest 
           | [] => Tree.NODE (NONE, [])

      val type_dict = buildTypePathDict refs
      val data_dict = buildDataPathDict refs

      fun indirectType path = 
          case Tree.find (sameElem, path, type_dict) of
            SOME _ => true
          | NONE => false

      fun indirectData path = 
          case Tree.find (sameElem, path, data_dict) of
            SOME _ => true
          | NONE => false

      val rec doBindings : dlbinding list * path -> dlbinding list

        = fn (binding :: rest, sofar) =>
             (case binding of
	        BINDING (n, t, d) =>
	        let
                  val sofar' = (Path.BINDING n) :: sofar
                  val here = rev sofar'
                  val d' = doData(d, sofar')
                in
                  if indirectData(here)
                  then BINDING (n, t, DATA_REF (ref d'))
                  else BINDING (n, t, d')
	       end

	     | TYPEDEF (n, t) =>
		let
		  val sofar' = (Path.BINDING n) :: sofar
                  val here = rev sofar'
		in
                  if indirectType(here)
                  then TYPEDEF (n, TYPE_REF (ref t))
                  else TYPEDEF (n, t)
                end)
             :: doBindings(rest, sofar)

           | ([], _) => []

      and doElements : int * dldata list * path -> dldata list

	= fn (i, elem :: rest, sofar) =>
	     let
	       val sofar' = (Path.INDEX i) :: sofar
               val here = rev sofar'
	       val elem' = doData(elem, sofar')
             in
               if indirectData(here)
               then DATA_REF (ref elem')
               else elem'
             end
             :: doElements(i+1, rest, sofar)

	   | (_, [], _) => []

      and doData : dldata * path -> dldata

	= fn (d, sofar) =>
	     case d of

	       DATA_REF dr => (dr := doData (!dr, sofar); d)

              (* No need to reverse bindings in this case. *)

	     | REC bindings => REC (doBindings (bindings, sofar))
                               
	     | VEC elements => VEC (doElements (0, elements, sofar))

	     | _ => d

    in
      doData (data, [])
    end


(* In the case of resolve, the stack is merely a list of bindings,
since we don't need to reconstruct the full path to the data. We just
need to get it. *)

local

  type binding_types = ((sym * dltype) list)
  type scope = dlbinding list
  type stack = scope list

  fun findDataInStack (path, bindings :: rest) =
      (case checkPath(path, REC bindings) of
        {tag=NOT_FOUND, ...} => findDataInStack (path, rest)
      | {tag=VALID_PATH, value = SOME (DATA v), ...} => SOME v
      | _ => NONE)
    | findDataInStack (_, []) = NONE

  fun findTypeInStack (path, bindings :: rest) =
      (case checkPath (path, REC bindings) of
         {tag=NOT_FOUND, ...} => findTypeInStack (path, rest)
       | {tag=VALID_PATH, value = SOME (TYPE t), ...} => SOME t
       | _ => NONE)
    | findTypeInStack (_, []) = NONE

in

val resolve : dldata -> dldata  = 
 fn data =>
    let
      val rec doBindingTypes : stack * binding_types -> binding_types
        = fn (stack, (name, t) :: rest) =>
             (name, doType(stack, t)) 
             :: doBindingTypes(stack, rest)
	   | (_, []) => []

      and doType : stack * dltype -> dltype
        = fn (stack, t) =>
             case t of
               REC_TYPE binding_types => 
	       REC_TYPE (doBindingTypes (stack, binding_types))
             | TYPE_PATH path =>
               (case findTypeInStack (path, stack) of 
                  SOME value => value
                | NONE => t) (* unresolved path *)
             | TYPE_REF tr => (tr := doType (stack, !tr); t)
             | VEC_TYPE t' => 
               VEC_TYPE (doType (stack, t'))
             | VECN_TYPE (n, t') => 
               VECN_TYPE (n, doType (stack, t'))
             | _ => t

      val rec doBindings : scope * stack * dlbinding list -> dlbinding list

        = fn (scope, stack, binding :: bindings) =>
             let
               val stack' = scope :: stack
               val binding' = 
                   case binding of
                     BINDING (n, t, d) 
                     => BINDING (n, 
                                 doType (stack', t), 
                                 doData (stack', d))

	           | TYPEDEF (n, t)
                     => (TYPEDEF (n, 
                                  doType (stack', t)))
               val scope' = binding' :: scope
             in
               binding' :: doBindings (scope', stack, bindings)
             end

	   | (_, _, []) => []

      and doElements : stack * dldata list -> dldata list

	= fn (stack, elem :: rest) =>
	     doData(stack, elem) :: doElements(stack, rest)
	   | (_, []) => []

      and doData : stack * dldata -> dldata

	= fn (stack, d) =>
	     case d of
	       DATA_PATH path =>
	       (case findDataInStack (path, stack) of
		  SOME value => value
		| NONE => d) (* unresolved ref *)
	     | DATA_REF dr => (dr := doData (stack, !dr); d)
	     | REC bindings => REC (rev (doBindings ([], stack, rev bindings)))
	     | VEC elements => VEC (doElements (stack, elements))
	     | _ => d
    in
      doData ([], data)
    end

end


(* 
   unresolve : dldata -> dldata 

   Only refs at bindings and vector elements are put in the
   dictionaries since only they can be addressed with a path.
*)

local

  type binding_types = ((sym * dltype) list)
  type dlbindings = dlbinding list

  (* paths are stored in reverse in the type ref and data ref dictionaries *)

  type dict = {t: (dltype ref * path) list,
               d: (dldata ref * path) list}

  val mergeDict : dict * dict -> dict
    = fn ({t=t1, d=d1}, {t=t2, d=d2}) => {t = t1 @ t2, 
                                          d = d1 @ d2}

  val addTypeRef : dict * dltype ref * path -> dict
    = fn ({t, d}, tr, p) => {t = (tr, p) :: t, d = d}

  val findTypePath : dict * dltype ref -> path option 
    = fn (dict, tr1) =>
         let
           val {t, ...} = dict
           fun find ([]) = NONE
             | find ((tr2, p) :: rest) =
               if tr1=tr2 then SOME p else find rest
         in
           find t
         end
  
  val addDataRef : dict * dldata ref * path -> dict
    = fn ({t, d}, dr, p) => {t = t, d = (dr, p) :: d}

  val findDataPath : dict * dldata ref -> path option 
    = fn (dict, dr1) =>
         let
           val {d, ...} = dict
           fun find ([]) = NONE
             | find ((dr2, p) :: rest) =
               if dr1=dr2 then SOME p else find rest
         in
           find d
         end
  

  val rec doBindingTypes : dict * binding_types * path -> binding_types
    = fn (dict, (n, t) :: rest, sofar) =>
         (n, doType(dict, t, sofar)) :: doBindingTypes (dict, rest, sofar)
       | (_, [], _) => []

  and doType : dict * dltype * path -> dltype 
    = fn (dict, t, sofar) =>
         case t of
           REC_TYPE binding_types => REC_TYPE (doBindingTypes (dict, binding_types, sofar))
         | TYPE_REF tr => 
           (case findTypePath (dict, tr) of
              SOME path => TYPE_PATH (Path.trim (rev sofar, rev path))
            | NONE => t)
         | VEC_TYPE t' => VEC_TYPE (doType (dict, t', sofar))
         | VECN_TYPE (n, t') => VECN_TYPE (n, (doType (dict, t', sofar)))
         | _ => t

  and doBindings : dict * dlbindings * path -> dlbindings * dict

    = fn (dict, binding :: bindings, sofar) =>
         (case binding of
            BINDING (n, t, d) =>
            let
              val sofar' = (Path.BINDING n) :: sofar
              val t' = doType (dict, t, sofar)
            in
              (case d of
                 DATA_REF dr =>
                 let
                   val path_opt = findDataPath (dict, dr)
                 in
                   case path_opt of
                     SOME path => 
                     let
                       val d' = DATA_PATH (Path.trim (rev sofar, rev path))
                       val (bindings', dict') = doBindings (dict, bindings, sofar)
                     in
                       ((BINDING (n, t', d')) :: bindings', dict')
                     end
                   | NONE => 
                     let 

(* I think the fact that this binding is not yet visible to
sub-bindings guarantees that a ref can only be resolved to something
bound in a previous scope. Vectors are trickier since a path must
always only begin with a symbol. So in the case of doElements,
separate dictionaries must be built and merged. If I've got my head
around the problem, this situation can only occur if someone builds a
data structure programmatically. It should never occurr from parsing a
data structure. *)

                       val (d', dict') = doData (dict, !dr, sofar')
                       val dict'' = addDataRef (dict', dr, sofar')
                       val (bindings', dict''') = doBindings (dict'', bindings, sofar)
                     in
                       ((BINDING (n, t', d')) :: bindings', dict''')
                     end
                 end
               | _ =>
                 let
                   val (d', dict') = doData(dict, d, sofar')
                   val (bindings', dict'') = doBindings (dict', bindings, sofar)
                 in
                   ((BINDING (n, t', d')) :: bindings', dict'')
                 end)
            end

          | TYPEDEF (n, t) =>
            let
              val sofar' = (Path.BINDING n) :: sofar
            in
              (case t of
                 TYPE_REF tr =>
                 let
                   val path_opt = findTypePath (dict, tr)
                 in
                   case path_opt of
                     SOME path => 
                     let
                       val t' = TYPE_PATH (Path.trim (rev sofar, rev path))
                       val (bindings', dict') = doBindings (dict, bindings, sofar)
                     in
                       ((TYPEDEF (n, t')) :: bindings', dict')
                     end
                   | NONE =>
                     let
                       val t' = doType (dict, !tr, sofar')
                       val dict' = addTypeRef (dict, tr, sofar')
                       val (bindings', dict'') = doBindings (dict', bindings, sofar)
                     in
                       ((TYPEDEF (n, t')) :: bindings', dict'')
                     end
                 end
               | _ =>
                 let
                   val t' = doType (dict, t, sofar')
                   val (bindings', dict') = doBindings (dict, bindings, sofar)
                 in
                   ((TYPEDEF (n, t')) :: bindings', dict')
                 end)
            end)

          | (dict, [], _) => ([], dict)

  and doElements : int * dict * dldata list * path -> dldata list * dict
    = fn (i, dict, d :: rest, sofar) =>
         let
           val sofar' = (Path.INDEX i) :: sofar
           val (d', dict_here) =
               (case d of 
                  DATA_REF dr =>
                  let
                    val path_opt = findDataPath (dict, dr)
                  in
                    case path_opt of
                      SOME path => (DATA_PATH (Path.trim (rev sofar, rev path)), dict)
                    | NONE => 
                      let
                        val (d'', dict'') = doData(dict, !dr, sofar')
                      in
                        (d'', addDataRef(dict'', dr, sofar'))
                      end
                  end
                | _ => doData(dict, d, sofar'))
           val (rest', dict_rest) = doElements(i+1, dict, rest, sofar)
         in
           (d' :: rest', mergeDict(dict_here, dict_rest))
         end
       | (_, dict, [], _) => ([], dict)

  and doData : dict * dldata * path -> dldata * dict
    = fn (dict, d, sofar) =>
         case d of
           DATA_REF dr => 
           (case findDataPath (dict, dr) of
              SOME path => (DATA_PATH (Path.trim (rev sofar, rev path)), dict)
            | NONE => (d, dict))
         | REC bindings => 
           let
             val (bindings', dict') = doBindings (dict, rev bindings, sofar)
           in
             (REC (rev bindings'), dict')
           end
         | VEC elements => 
           let
             val (elements', dict') = doElements (0, dict, elements, sofar)
           in
             (VEC elements', dict')
           end
         | _ => (d, dict)

in

fun unresolve data = 
    let
      val (data', _) = doData({t=[],d=[]}, data, [])
    in
      data'
    end

end



(* The 'flatten' function maps binding types of the record passed in
down the structures. The binding and item helper functions take
'unwrapped' types. *)


fun flattenRecord (REC bindings) =
    let
      fun flattenBinding (b, []) = b
        | flattenBinding (TYPEDEF td, _) = (TYPEDEF td)
        | flattenBinding (BINDING (a_s, a_t, d), (b_s, b_t) :: rest) = 
          (if sameSym(a_s,b_s)
           then 
             let
               val t = specificType(a_t, b_t)
             in
               (BINDING (a_s, t, flattenData(d, t)))
             end
           else
             flattenBinding (BINDING (a_s, a_t, d), rest))

      and flattenBindings ([], _) = []
        | flattenBindings ((b :: rest), bts) =
          (flattenBinding (b, bts)) :: flattenBindings(rest, bts)

      and flattenItems ([], _) = []
        | flattenItems (i :: rest, t) =
          (flattenData (i, t) :: flattenItems (rest, t))

      and flattenData (REC bindings, REC_TYPE bts) 
        = REC (flattenBindings (bindings, bts))

        | flattenData (VEC items, VEC_TYPE t) 
          = VEC (flattenItems (items, t))

        | flattenData (VEC items, VECN_TYPE (_, t)) 
          = VEC (flattenItems (items, t))

        | flattenData (DATA_REF (r as (ref d)), t)
          = (r := flattenData(d,t); DATA_REF r)

        | flattenData (d, TYPE_REF (ref t))
          = flattenData(d, t)

        | flattenData (d, _) = d

      fun doit ([]) = []

        | doit (BINDING (s, t, d) :: rest)
          = BINDING (s, t, flattenData (d, t)) :: doit rest

        | doit (TYPEDEF td :: rest)
          = TYPEDEF td :: doit (rest)
    in

      REC (doit bindings)

    end

  | flattenRecord (d) = d


fun map f d = 
    case d of 

      DATA_REF (r as (ref d)) => (r := map f d; f (DATA_REF r))

    | REC bindings =>
      let
        fun doit (BINDING (s, t, d) :: rest)
          = (BINDING (s, t, map f d) :: doit rest)
          | doit (TYPEDEF td :: rest) = (TYPEDEF td :: doit rest)
          | doit [] = []
      in
        (f (REC (doit bindings)))
      end

    | VEC items =>
      let
        fun doit (d :: rest) = (map f d :: doit rest)
          | doit [] = []
      in
        (f (VEC (doit items)))
      end

    | d => f d

fun flatten d = map flattenRecord d

end (* struct DL *)

